from sparknlp.internal import ExtendedJavaWrapper
from pyspark.sql import SparkSession, DataFrame


class OcrHelper(ExtendedJavaWrapper):

    def __init__(self):
        super(OcrHelper, self).__init__("com.johnsnowlabs.nlp.util.io.OcrHelper")

    def createDataset(self, spark, input_path):
        if type(spark) != SparkSession:
            raise Exception("spark must be SparkSession")
        return DataFrame(self._java_obj.createDataset(spark._jsparkSession, input_path), spark)

    def createMap(self, input_path):
        return self._java_obj.createMap(input_path)

    def setPreferredMethod(self, value):
        return self._java_obj.setPreferredMethod(value)

    def getPreferredMethod(self):
        return self._java_obj.getPreferredMethod()

    def setFallbackMethod(self, value):
        return self._java_obj.setFallbackMethod(value)

    def getFallbackMethod(self):
        return self._java_obj.getFallbackMethod()

    def setMinSizeBeforeFallback(self, value):
        return self._java_obj.setMinSizeBeforeFallback(value)

    def getMinSizeBeforeFallback(self):
        return self._java_obj.getMinSizeBeforeFallback()

    def setEngineMode(self, mode):
        return self._java_obj.setEngineMode(mode)

    def getEngineMode(self):
        return self._java_obj.getEngineMode()

    def setPageSegMode(self, mode):
        return self._java_obj.setPageSegMode(mode)

    def getPageSegMode(self):
        return self._java_obj.getPageSegMode()

    def setPageIteratorLevel(self, level):
        return self._java_obj.setPageIteratorLevel(level)

    def getPageIteratorLevel(self):
        return self._java_obj.getPageIteratorLevel()

    def setScalingFactor(self, factor):
        return self._java_obj.setScalingFactor(factor)

    def setSplitPages(self, value):
        return self._java_obj.setSplitPages(value)

    def getSplitPages(self):
        return self._java_obj.getSplitPages()

    def setSplitRegions(self, value):
        return self._java_obj.setSplitRegions(value)

    def getSplitRegions(self):
        return self._java_obj.getSplitRegions()

    def useErosion(self, use, k_size=2, k_shape=0):
        return self._java_obj.useErosion(use, k_size, k_shape)

    def drawRectanglesToFile(self, path, coordinates, output_path):
        if type(coordinates) != list or len(coordinates) == 0:
            raise Exception("coordinates not a list, or is empty")
        if type(coordinates[0]) == Coordinate:
            jcoords = list(map(lambda c: c.java_obj, coordinates))
        else:
            jcoords = list(map(lambda c: Coordinate(c['i'], c['p'], c['x'], c['y'], c['w'], c['h']).java_obj, coordinates))
        return self._java_obj.drawRectanglesToFile(path, jcoords, output_path)

    def drawRectanglesDataset(
            self,
            spark,
            dataset,
            filename_col='filename',
            pagenum_col='pagenum',
            coordinates_col='coordinates',
            output_location='./highlighted/',
            output_suffix='_draw'):
        jspark = spark._jsparkSession
        jdf = dataset._jdf
        return self._java_obj.drawRectanglesDataset(jspark, jdf, filename_col, pagenum_col, coordinates_col, output_location, output_suffix)


#
# @param i  Chunk index.
# @param p  Page number.
# @param x  The lower left x coordinate.
# @param y  The lower left y coordinate.
# @param w  The width of the rectangle.
# @param h  The height of the rectangle.
#
class Coordinate(ExtendedJavaWrapper):
    def __init__(self, i, p, x, y, w, h):
        super(Coordinate, self).__init__("com.johnsnowlabs.nlp.util.io.schema.Coordinate", i, p, x, y, w, h)
        self.i = i
        self.p = p
        self.x = x
        self.y = y
        self.w = w
        self.h = h

    def __str__(self):
        return "Coordinate(i=%s, p=%s, x=%s, y=%s, w=%s, h=%s)" % (self.i, self.p, self.x, self.y, self.w, self.h)



from sparknlp.internal import ExtendedJavaWrapper


class NorvigSpellEvaluation(ExtendedJavaWrapper):

    def __init__(self, spark, test_file, ground_truth_file):
        ExtendedJavaWrapper.__init__(self, "com.johnsnowlabs.nlp.eval.spell.NorvigSpellEvaluation", spark._jsparkSession, test_file, ground_truth_file)

    def computeAccuracyAnnotator(self, train_file, spell):
        input_cols = spell.getInputCols()
        output_col = spell.getOutputCol()
        java_input_cols = self.new_java_array_string(input_cols)
        return self._java_obj.computeAccuracyAnnotator(train_file, java_input_cols, output_col, spell.dictionary_path)

    def computeAccuracyModel(self, spell):
        return self._java_obj.computeAccuracyModel(spell._java_obj)


class SymSpellEvaluation(ExtendedJavaWrapper):

    def __init__(self, spark, test_file, ground_truth_file):
        ExtendedJavaWrapper.__init__(self, "com.johnsnowlabs.nlp.eval.spell.SymSpellEvaluation", spark._jsparkSession, test_file, ground_truth_file)

    def computeAccuracyAnnotator(self, train_file, spell):
        input_cols = spell.getInputCols()
        output_col = spell.getOutputCol()
        java_input_cols = self.new_java_array_string(input_cols)
        return self._java_obj.computeAccuracyAnnotator(train_file, java_input_cols, output_col, spell.dictionary_path)

    def computeAccuracyModel(self, spell):
        return self._java_obj.computeAccuracyModel(spell._java_obj)


class NerDLEvaluation(ExtendedJavaWrapper):

    def __init__(self, spark, test_file, tag_level=""):
        ExtendedJavaWrapper.__init__(self, "com.johnsnowlabs.nlp.eval.ner.NerDLEvaluation", spark._jsparkSession, test_file, tag_level)

    def computeAccuracyModel(self, ner):
        return self._java_obj.computeAccuracyModel(ner._java_obj)

    def computeAccuracyAnnotator(self, train_file, ner, embeddings):
        ner_params = self.__getNerParams(ner)
        embeddings_params = self.__getEmbeddingsParams(embeddings)
        return self._java_obj.computeAccuracyAnnotator(train_file, ner_params['input_cols'], ner_params['output_col'],
                                                       ner_params['random_seed'], ner_params['label_column'],
                                                       embeddings_params['input_cols'], embeddings_params['output_col'],
                                                       embeddings_params['path'], embeddings_params['dimension'],
                                                       embeddings_params['format'])

    def __getNerParams(self, ner):
        ner_params = dict()
        input_cols = ner.getInputCols()
        ner_params['input_cols'] = self.new_java_array_string(input_cols)
        ner_params['output_col'] = ner.getOutputCol()
        ner_params['label_column'] = ner.getLabelColumn()
        ner_params['random_seed'] = ner.getRandomSeed()
        return ner_params

    def __getEmbeddingsParams(self, embeddings):
        embeddings_params = dict()
        input_cols = embeddings.getInputCols()
        embeddings_params['input_cols'] = self.new_java_array_string(input_cols)
        embeddings_params['output_col'] = embeddings.getOutputCol()
        embeddings_params['path'] = embeddings.getEmbeddingsPath()
        embeddings_params['dimension'] = embeddings.getDimension()
        embeddings_params['format'] = embeddings.getFormat()
        return embeddings_params


class NerCrfEvaluation(ExtendedJavaWrapper):

    def __init__(self, spark, test_file, tag_level=""):
        ExtendedJavaWrapper.__init__(self, "com.johnsnowlabs.nlp.eval.ner.NerCrfEvaluation", spark._jsparkSession, test_file, tag_level)

    def computeAccuracyModel(self, ner):
        return self._java_obj.computeAccuracyModel(ner._java_obj)

    def computeAccuracyAnnotator(self, train_file, ner, embeddings):
        ner_params = self.__getNerParams(ner)
        embeddings_params = self.__getEmbeddingsParams(embeddings)
        return self._java_obj.computeAccuracyAnnotator(train_file, ner_params['input_cols'], ner_params['output_col'],
                                                       ner_params['random_seed'], ner_params['label_column'],
                                                       embeddings_params['input_cols'], embeddings_params['output_col'],
                                                       embeddings_params['path'], embeddings_params['dimension'],
                                                       embeddings_params['format'])

    def __getNerParams(self, ner):
        ner_params = dict()
        input_cols = ner.getInputCols()
        ner_params['input_cols'] = self.new_java_array_string(input_cols)
        ner_params['output_col'] = ner.getOutputCol()
        ner_params['label_column'] = ner.getLabelColumn()
        ner_params['random_seed'] = ner.getRandomSeed()
        return ner_params

    def __getEmbeddingsParams(self, embeddings):
        embeddings_params = dict()
        input_cols = embeddings.getInputCols()
        embeddings_params['input_cols'] = self.new_java_array_string(input_cols)
        embeddings_params['output_col'] = embeddings.getOutputCol()
        embeddings_params['path'] = embeddings.getEmbeddingsPath()
        embeddings_params['dimension'] = embeddings.getDimension()
        embeddings_params['format'] = embeddings.getFormat()
        return embeddings_params


class POSEvaluation(ExtendedJavaWrapper):

    def __init__(self, spark, test_file):
        ExtendedJavaWrapper.__init__(self, "com.johnsnowlabs.nlp.eval.POSEvaluation", spark._jsparkSession, test_file)

    def computeAccuracyModel(self, pos):
        return self._java_obj.computeAccuracyModel(pos._java_obj)

    def computeAccuracyAnnotator(self, train_file, pos):
        pos_params = self.__getPosParams(pos)
        return self._java_obj.computeAccuracyAnnotator(train_file, pos_params['input_cols'], pos_params['output_col'],
                                                       pos_params['number_iterations'])

    def __getPosParams(self, pos):
        pos_params = dict()
        input_cols = pos.getInputCols()
        pos_params['input_cols'] = self.new_java_array_string(input_cols)
        pos_params['output_col'] = pos.getOutputCol()
        pos_params['number_iterations'] = pos.getNIterations()
        return pos_params



from pyspark.ml.util import JavaMLWritable
from pyspark.ml.wrapper import JavaModel, JavaEstimator
from pyspark.ml.param.shared import Param, TypeConverters
from sparknlp.util import AnnotatorJavaMLReadable
from pyspark.ml.param import Params
from pyspark import keyword_only
import sparknlp.internal as _internal
import re


class AnnotatorProperties(Params):

    inputCols = Param(Params._dummy(),
                      "inputCols",
                      "previous annotations columns, if renamed",
                      typeConverter=TypeConverters.toListString)
    outputCol = Param(Params._dummy(),
                      "outputCol",
                      "output annotation column. can be left default.",
                      typeConverter=TypeConverters.toString)

    def setInputCols(self, value):
        return self._set(inputCols=value)

    def setOutputCol(self, value):
        return self._set(outputCol=value)


# Helper class used to generate the getters for all params
class ParamsGettersSetters(Params):
    getter_attrs = []

    def __init__(self):
        super(ParamsGettersSetters, self).__init__()
        for param in self.params:
            param_name = param.name
            fg_attr = "get" + re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), param_name)
            fs_attr = "set" + re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), param_name)
            # Generates getter and setter only if not exists
            try:
                getattr(self, fg_attr)
            except AttributeError:
                setattr(self, fg_attr, self.getParamValue(param_name))
            try:
                getattr(self, fs_attr)
            except AttributeError:
                setattr(self, fs_attr, self.setParamValue(param_name))

    def getParamValue(self, paramName):
        def r():
            try:
                return self.getOrDefault(paramName)
            except KeyError:
                return None
        return r

    def setParamValue(self, paramName):
        def r(v):
            self.set(self.getParam(paramName), v)
            return self
        return r


class AnnotatorModel(JavaModel, AnnotatorJavaMLReadable, JavaMLWritable, AnnotatorProperties, ParamsGettersSetters):

    @keyword_only
    def setParams(self):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    @keyword_only
    def __init__(self, classname, java_model=None):
        super(AnnotatorModel, self).__init__(java_model=java_model)
        if classname and not java_model:
            self.__class__._java_class_name = classname
            self._java_obj = self._new_java_obj(classname, self.uid)
        if java_model is not None:
            self._transfer_params_from_java()


class HasEmbeddings(Params):
    dimension = Param(Params._dummy(),
                      "dimension",
                      "Number of embedding dimensions",
                      typeConverter=TypeConverters.toInt)

    caseSensitive = Param(Params._dummy(),
                                "caseSensitive",
                                "whether to ignore case in tokens for embeddings matching",
                                typeConverter=TypeConverters.toBoolean)

    def setDimension(self, value):
        return self._set(dimension=value)

    def setCaseSensitive(self, value):
        return self._set(caseSensitive=value)


class HasWordEmbeddings(HasEmbeddings):
    embeddingsRef = Param(Params._dummy(),
                          "embeddingsRef",
                          "if sourceEmbeddingsPath was provided, name them with this ref. Otherwise, use embeddings by this ref",
                          typeConverter=TypeConverters.toString)

    includeEmbeddings = Param(Params._dummy(),
                           "includeEmbeddings",
                           "whether or not to save indexed embeddings along this annotator",
                           typeConverter=TypeConverters.toBoolean)

    def setEmbeddingsRef(self, value):
        from sparknlp.annotator import WordEmbeddingsModel
        if type(self) == WordEmbeddingsModel and self.getParam('embeddingsRef'):
            raise Exception("Cannot override embeddings ref on a WordEmbeddingsModel. Please re-use current ref: %s" % self.getOrDefault('embeddingsRef'))
        return self._set(embeddingsRef=value)

    def getEmbeddingsRef(self):
        return self.getOrDefault('embeddingsRef')

    def setIncludeEmbeddings(self, value):
        return self._set(includeEmbeddings=value)

    def getIncludeEmbeddings(self):
        return self.getOrDefault("includeEmbeddings")


class AnnotatorApproach(JavaEstimator, JavaMLWritable, AnnotatorJavaMLReadable, AnnotatorProperties,
                        ParamsGettersSetters):

    trainingCols = Param(Params._dummy(),
                               "trainingCols",
                               "the training annotation columns. uses input annotation columns if missing",
                               typeConverter=TypeConverters.toListString)

    @keyword_only
    def __init__(self, classname):
        ParamsGettersSetters.__init__(self)
        self.__class__._java_class_name = classname
        self._java_obj = self._new_java_obj(classname, self.uid)

    def setTrainingCols(self, cols):
        return self._set(trainingCols=cols)


def RegexRule(rule, identifier):
    return _internal._RegexRule(rule, identifier).apply()


class ReadAs(object):
    LINE_BY_LINE = "LINE_BY_LINE"
    SPARK_DATASET = "SPARK_DATASET"


def ExternalResource(path, read_as=ReadAs.LINE_BY_LINE, options={}):
    return _internal._ExternalResource(path, read_as, options).apply()
